// file: life.cpp

#include <stdlib.h>
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <iostream>
#include <new.h>
#include "life.h"
#include "lifeconstants.h"


/*  Memory for 'cells', 'checked' and 'trail' tables is now allocated
    dynamically. Unfortenately, I didn't get 'memset' working correctly
    with the memory blocks allocated by 'new []'-operator. */ 

Life::Life(int c, int r)
{
    if(r <= MAXROWS && r >= MINROWS) 
	rows = r;
    else 
	rows = (r < MINROWS) ? MINROWS : MAXROWS;
    if(c <= MAXCOLS && c >= MINCOLS)
	cols = c;
    else 
	cols = (c < MINCOLS) ? MINCOLS : MAXCOLS;
    start = new Cell(0,0);
    trailing = false;
    age = 0;
    population = 0;
    pSurface = &Life::torus;

    try {
	cells = new bool * [rows+2];
	for(int i = 0; i < rows+2; i++) 
	    cells[i] = new bool[cols+2]; 

	checked = new bool * [rows+2];
	for(int i = 0; i < rows+2; i++) 
	    checked[i] = new bool[cols+2]; 

	trail = new bool * [rows+2];
	for(int i = 0; i < rows+2; i++) 
	    trail[i] = new bool[cols+2]; 

    } catch (bad_alloc) {
	std::cerr << "Memory allocation failed!\n";
	exit(1);
    }
 
    for(int i = 0; i < rows+2; i++) {
	for(int j = 0; j < cols+2; j++) {
	    cells[i][j] = false;
	    checked[i][j] = false;
	    trail[i][j] = false;
	}
    }
}

Life::~Life() 
{
    for(int i = 0; i < rows+2; i++) {
	delete[] cells[i];
    } 
    delete[] cells;

    for(int i = 0; i < rows+2; i++) {
	delete[] checked[i];
    } 
    delete[] checked;

    for(int i = 0; i < rows+2; i++) {
	delete[] trail[i];
    } 
    delete[] trail;
}

/* Deletes a list of cells starting at 'first'  */

void Life::deleteList(Cell * first)
{
    if(first == 0)
	return;
    Cell * c = first;
    Cell * prev;
    while(c != 0) {
	prev = c;
	c = c->next;
	delete prev;
    }
}

/* Both the table and the list are cleared. */

void Life::clear()
{
    for ( int i = 0; i < rows + 2; i++ ) {
	for ( int j = 0; j < cols + 2; j++ ) {
	    cells[i][j] = false;
	}
    }
    updateList();
    age = 0;
}

void Life::hSymmetrize() 
{
    Cell * c = start->next;
    while(c != 0) {
	cells[c->row][cols - c->col + 1] = true;
	c = c->next;
    }
    updateList();
}

void Life::vSymmetrize() 
{
    Cell * c = start->next;
    while(c != 0) {
	cells[rows - c->row + 1][c->col] = true;
	c = c->next;
    }
    updateList();
}


bool Life::pdSymmetrize()
{
    int maxRow = getMaxRow(start->next);
    int maxCol = getMaxCol(start->next);

    if(maxRow > cols || maxCol > rows)
	return false;

    Cell * c = start->next;
    while(c != 0) {
	cells[c->col][c->row] = true;
	c = c->next;
    }
    updateList();
    return true;

}


bool Life::mdSymmetrize()
{
    int minRow = getMinRow(start->next);
    int minCol = getMinCol(start->next);

    if(minRow <= rows - cols || minCol <= cols - rows)
	return false;

    Cell * c = start->next;
    while(c != 0) {
	cells[cols - c->col + 1][rows - c->row + 1] = true;
	c = c->next;
    }
    updateList();
    return true;

}


bool Life::r90Transform()
{
    int minCol = getMinCol(start->next);
    int maxRow = getMaxRow(start->next);
    if(minCol <= cols - rows || maxRow > cols)
	return false;
    Cell * c = start->next;
    while(c != 0) {
	cells[cols - c->col + 1][c->row] = true; 
	c = c->next;
    }
    updateList();
    return true;
} 


void Life::r180Symmetrize()
{
    Cell * c = start->next;
    while(c != 0) {
	cells[rows - c->row + 1][cols -c->col + 1] = true; 
	c = c->next;
    }
    updateList();
} 


bool Life::r270Transform()
{
    int maxCol = getMaxCol(start->next);
    int minRow = getMinRow(start->next);
    if(maxCol > rows || minRow <= rows - cols)
	return false;
 
    Cell * c = start->next;
    while(c != 0) {
	cells[c->col][rows - c->row + 1] = true; 
	c = c->next;
    }
    updateList();
    return true;
} 


int Life::getMinRow(Cell * first)
{  
    if(first == 0)
	return -1;
    int minRow = this->rows;
    Cell * c = first;
    while(c != 0) {
	minRow = c->row < minRow ? c->row : minRow;
	c = c->next;
    }
    return minRow;
}


int Life::getMinCol(Cell * first)
{  
    if(first == 0)
	return -1;
    int minCol = this->cols;
    Cell * c = first;
    while(c != 0) {
	minCol = c->col < minCol ? c->col : minCol;
	c = c->next;
    }
    return minCol;
}


int Life::getMaxRow(Cell * first)
{  
    if(first == 0)
	return -1;
    int maxRow = 1;
    Cell * c = first;
    while(c != 0) {
	maxRow = c->row > maxRow ? c->row : maxRow;
	c = c->next;
    }
    return maxRow;
}



int Life::getMaxCol(Cell * first)
{  
    if(first == 0)
	return -1;
    int maxCol = 1;
    Cell * c = first;
    while(c != 0) {
	maxCol = c->col > maxCol ? c->col : maxCol;
	c = c->next;
    }
    return maxCol;
}


void Life::stepUp(Cell * c)
{
    if(c->row > 1) 
	c->row -= 1;
    else if(c->row == 1) 
	c->row = rows;
}

void Life::stepUp() 
{
    Cell * prevStart = copyCellList(start);
    Cell * c = start->next;
    while(c != 0) {
	stepUp(c);
	c = c->next;
    }
    setCells(prevStart);    
    deleteList(prevStart);
}


void Life::stepDown(Cell * c)
{
    if(c->row < rows) 
	c->row += 1;
    else if(c->row == rows)
	c->row = 1;
}



void Life::stepDown() 
{
    Cell * prevStart = copyCellList(start);
    Cell * c = start->next;
    while(c != 0) {
	stepDown(c);
	c = c->next;
    }
    setCells(prevStart);    
    deleteList(prevStart);
}


void Life::stepLeft(Cell * c)
{
    if(c->col > 1) 
	c->col -= 1;
    else if(c->col == 1)
	c->col = cols;
}

void Life::stepLeft() {
    Cell * prevStart = copyCellList(start);
    Cell * c = start->next;
    while(c != 0) {
	stepLeft(c);
	c = c->next;
    }
    setCells(prevStart);    
    deleteList(prevStart);
}


void Life::stepRight(Cell * c)
{
    if(c->col < cols) 
	c->col += 1;
    else if(c->col == cols) 
	c->col = 1;
}


void Life::stepRight()
{ 
    Cell * prevStart = copyCellList(start);
    Cell * c = start->next;
    while(c != 0) {
	stepRight(c);
	c = c->next;
    }
    setCells(prevStart);    
    deleteList(prevStart);
}



/* This function updates the list of live cells. It's called after 
   'cells'-table has changed or simply to initialize the list   */

void Life::updateList() 
{
    population = 0;
    Cell * initStart = new Cell(0,0);
    Cell * cell, * prior = initStart;
    for(int i = 1; i < rows+1; i++) {
        for(int j = 1; j < cols+1; j++) {
	    if(cells[i][j]) {
		cell = new Cell(i,j);
		prior->next = cell;
		prior = cell;
		population++;
	    }
	}
    }
    (this->*pSurface)();
    deleteList(start); 
    start = initStart;
}

/* This is the opposite of the preceding function. First every cell 
of the table is set to 'false' via a call to 'memset' and then those 
cells that are in the list are set to 'true'.   */  


void Life::setCells(Cell * prevStart) {
    population = 0;
    Cell * cell = prevStart->next;
    while(cell != 0) {
	cells[cell->row][cell->col] = false;
	cell = cell->next;
    }
    if(start != 0)
	cell = start->next;
    while(cell != 0) {
	cells[cell->row][cell->col] = true;
	population++;
	cell = cell->next;
    }
    (this->*pSurface)();
}

/* Counts live neighbours of the cell cells[h][w],where h and w must
   satisfy  0 < h < rows+1 and 0 < w < cols+1.  */

int Life::countNeighbours(int h, int w) {
    int sum = 0;		
    for(int i = h-1; i < h+2; i++) {
	for(int j = w-1; j < w+2; j++) {
	    sum += cells[i][j];
	}
    }
    return (sum - cells[h][w]);
}


/* Checks whether a cell corresponding to 'c' or some of its 8 neighbours 
  should be alive in the next generation. If such a cell is found a new 
  'Cell' object is created and pushed into the end of the list whose last 
  item is 'prior'. For all 9 cells the corresponding entry in the 
  'checked'-table is set to 'true' to prevent from multiple checking. 
  Only those cells with 'checked'-table entry set to 'false' are checked, 
  of course.    
*/

void Life::checkCell(Cell * c) {
    int ln, i, j;
    Cell * cell;
    for(i = c->row-1; i < c->row+2; i++) {
	for(j = c->col-1; j < c->col+2; j++) {
	    if(!checked[i][j]) {
		checked[i][j] = true;
		ln = countNeighbours(i,j);
		if(ln == 3 || ln + cells[i][j] == 3) {
		    cell = new Cell(i,j); 
		    prior->next = cell;   
		    prior = cell;
		}
	    }
	}
    }
}

/* Creates the next generation. First every 'checked' entry is set
to 'false' and then the entries of 2 outermost layers are set to 
'true' to prevent 'countNeighbours' to be called with invalid values. 
Next the cells in the list of living cells of the current 
generation are checked. Then the outermost layer of visible entries 
of 'cells'-table are checked without resorting to 'checkCell'. 
Then the old list is deleted and the list of live cells of 
the next generation is linked to 'start'. Finally 'setCells' is 
called to update 'cells'-table. 
*/

void Life::nextGen()
{
    int i, j, ln;
    Cell * c = 0, * newStart; 
    if(trailing) 
	doTrail();

    for(int i = 0; i < rows+2; i++) {
	for(int j = 0; j < cols+2; j++) {
	    checked[i][j] = false;
	}
    }

    for(i = 0; i < rows+2; i++) {
	checked[i][0] = checked[i][1] =                                                 checked[i][cols] = checked[i][cols+1] = true;
	}
    
    for(j = 2; j < cols; j++) {
	checked[0][j] = checked[1][j] =                                                 checked[rows][j] = checked[rows+1][j] = true;
    }

    newStart = new Cell(0,0);
    prior = newStart;

    if(start != 0)
	c = start->next;
    while(c != 0) {
	checkCell(c);
	c = c->next;
    }

    for(i = 1; i < rows+1; i++) {
	ln = countNeighbours(i, 1);
	if(ln == 3 || ln + cells[i][1] == 3) {
	    c = new Cell(i, 1);
	    prior->next = c;
	    prior = c;
	}
    }

    for(i = 1; i < rows+1; i++) {
	ln = countNeighbours(i,cols);
	if(ln == 3 || ln + cells[i][cols] == 3) {
	    c = new Cell(i,cols);
	    prior->next = c;
	    prior = c;
	}
    }

    for(j = 2; j < cols; j++) {
	ln = countNeighbours(1,j);
	if(ln == 3 || ln + cells[1][j] == 3) {
	    c = new Cell(1,j);
	    prior->next = c;
	    prior = c;
	}
    }


    for(j = 2; j < cols; j++) {
	ln = countNeighbours(rows,j);
	if(ln == 3 || ln + cells[rows][j] == 3) {
	    c = new Cell(rows,j);
	    prior->next = c;
	    prior = c;
	}
    }

    Cell * prevStart = start;
    start = newStart;
    setCells(prevStart);
    deleteList(prevStart);
    age++;

}



void Life::randomSeed()
{
    population = 0;
    srand(time(0));
    for(int i = 1; i <= rows; i++) {
	for(int j = 1; j <= cols; j++) {
	    if(cells[i][j] =                                                                     (((double)rand())/((double)RAND_MAX) > 0.5))
		population++;
	}
    }
    updateList();
    age = 0;

}


Cell * Life::toList()
{
    Cell * c, * last = 0, * start = 0;;
    for(int i = 1; i < rows+1; i++) {
	for(int j = 1; j < cols+1; j++) {
	    if(cells[i][j]) {
		if(start != 0) {
		    c = new Cell(i, j);
		    last->next = c;
		    last = c;
		}
		else {
		    start = new Cell(i, j);
		    last = start;
		}
	    }
	}
    }
    return start; 
}

/* List of cells, starting at 'first', is loaded. */ 

bool Life::loadList(Cell * first)
{
    bool tooBig = false;
    int minCol = getMinCol(first);
    int maxCol = getMaxCol(first);
    int minRow = getMinRow(first);
    int maxRow = getMaxRow(first);
    int h = maxRow - minRow + 1;
    int w = maxCol - minCol + 1;

    if(h <= rows && w <= cols) {
	int rowShift = (rows - h)/2 - minRow + 1;
	int colShift = (cols - w)/2 - minCol + 1;
	Cell * c = first;  
	while(c != 0) {
	    cells[c->row + rowShift][c->col + colShift] = true;
	    c = c->next;
	}
    }
    else 
	tooBig = true;

    return tooBig; 
}



void Life::torus()
{
    int i;

    cells[0][0] = cells[rows][cols];
    cells[0][cols+1] = cells[rows][1];
    cells[rows+1][0] = cells[1][cols];
    cells[rows+1][cols+1] = cells[1][1];

    for(i = 1; i < cols+1; i++) {
	cells[rows+1][i] = cells[1][i];
	cells[0][i] = cells[rows][i];
    }

    for(i = 1; i < rows+1; i++) {
	cells[i][0] = cells[i][cols];
	cells[i][cols+1] = cells[i][1];
    }
}

void Life::ppp()
{
    int i;

    cells[0][0] = cells[1][1];
    cells[0][cols+1] = cells[1][cols];
    cells[rows+1][0] = cells[rows][1];
    cells[rows+1][cols+1] = cells[rows][cols];

    for(i = 1; i < cols+1; i++) {
	cells[rows+1][i] = cells[1][cols+1-i];
	cells[0][i] = cells[rows][cols+1-i];
    }
    for(i = 1; i < rows+1; i++) {
	cells[i][0] = cells[rows+1-i][cols];
	cells[i][cols+1] = cells[rows+1-i][1];
    }

}


void Life::kleinBottle()
{
    int i;

    cells[0][0] = cells[rows][1];
    cells[0][cols+1] = cells[rows][cols];
    cells[rows+1][0] = cells[1][1];
    cells[rows+1][cols+1] = cells[1][cols];


    for(i = 1; i < cols+1; i++) {
	cells[rows+1][i] = cells[1][cols+1-i];
	cells[0][i] = cells[rows][cols+1-i];
    }

    for(i = 1; i < rows+1; i++) {
	cells[i][0] = cells[i][cols];
	cells[i][cols+1] = cells[i][1];
    }

}


void Life::mirror()
{
    cells[0][0] = cells[0][cols+1] =                                                              cells[rows+1][0] = cells[rows+1][cols+1] = false; 
    int i;
    for(i = 1; i < cols + 1; i++) {
	cells[0][i] = cells[1][i];
	cells[rows+1][i] = cells[rows][i];
    }

    for(i = 1; i < rows + 1; i++) {
	cells[i][0] = cells[i][1];
	cells[i][cols+1] = cells[i][cols];
    }
} 


void Life::box()
{
    cells[0][0] = cells[0][cols+1] =                                                              cells[rows+1][0] = cells[rows+1][cols+1] = false; 
    int i;
    for(i = 1; i < cols + 1; i++) {
	cells[0][i] = false;
	cells[rows+1][i] = false;
    }

    for(i = 1; i < rows + 1; i++) {
	cells[i][0] = false;
	cells[i][cols+1] = false;
    }

}


void Life::createBBC()
{
    if(population < 2)
	return;

    int minCol = getMinCol(start->next);
    int maxCol = getMaxCol(start->next);
    int minRow = getMinRow(start->next);
    int maxRow = getMaxRow(start->next); 

    for(int i = minRow; i <= maxRow; i++) {
	for(int j = minCol; j <= maxCol; j++) {
	    cells[i][j] = !cells[i][j];
	}
    }

    updateList();
    age = 0;

}

void Life::createSlim()
{
    if(population < 9)
	return; 
    Cell * inn = new Cell(0,0);
    Cell * prev = inn, * cell;
    Cell * c = start->next;
    while(c != 0) {
	if(countNeighbours(c->row, c->col) == 8) {
	    cell = new Cell(c->row, c->col);
	    prev->next = cell;
	    prev = cell;
	}
	c = c->next;
    }

    removeBorder();

    cell = inn->next;
    while(cell != 0) {
	cells[cell->row][cell->col] = 0;
	cell = cell->next;
    }

    updateList();
    age = 0;
    deleteList(inn);
}

void Life::removeBorder()
{
    Cell * removeStart = new Cell(0, 0);
    Cell * c, * removeLast = removeStart;
    int i, j;
    int minCol = getMinCol(start->next);
    int maxCol = getMaxCol(start->next);
    int minRow = getMinRow(start->next);
    int maxRow = getMaxRow(start->next); 

    for(i = minRow+1; i < maxRow; i++) {
	if(countNeighbours(i, maxCol) == 5) {
	    c = new Cell(i, maxCol);
	    removeLast->next = c;
	    removeLast = c;
	}

	if(countNeighbours(i, minCol) == 5) {
	    c = new Cell(i, minCol);
	    removeLast->next = c;
	    removeLast = c;
	}
    }

    for(j = minCol+1; j < maxCol; j++) {
	if(countNeighbours(maxRow, j) == 5) {
	    c = new Cell(maxRow, j);
	    removeLast->next = c;
	    removeLast = c;
	}

	if(countNeighbours(minRow, j) == 5) {
	    c = new Cell(minRow, j);
	    removeLast->next = c;
	    removeLast = c;
	}
    }

    if(countNeighbours(minRow, minCol) == 3) {
	c = new Cell(minRow, minCol);
	removeLast->next = c;
	removeLast = c;
    }

    if(countNeighbours(minRow, maxCol) == 3) {
	c = new Cell(minRow, maxCol);
	removeLast->next = c;
	removeLast = c;
    }

    if(countNeighbours(maxRow, minCol) == 3) {
	c = new Cell(maxRow, minCol);
	removeLast->next = c;
	removeLast = c;
    }

    if(countNeighbours(maxRow, maxCol) == 3) {
	c = new Cell(maxRow, maxCol);
	removeLast->next = c;
    }

    c = removeStart->next;
    while(c != 0) {
	cells[c->row][c->col] = false;
	c = c->next;
    }

    deleteList(removeStart);
}



void Life::doTrail() {
    if(start == 0)
	return;
    Cell * cell = start->next;
	while(cell != 0) {
	    trail[cell->row][cell->col] = true;
	    cell = cell->next;
	}
    }


Cell * Life::copyCellList(Cell * first)
{
    Cell * copyStart = new Cell(first->row, first->col);
    Cell * copyLast = copyStart;
    Cell * cell, * c = first->next;
    while(c != 0) {
	cell = new Cell(c->row, c->col);
	copyLast->next = cell;
	copyLast = cell;
	c = c->next; 
    }
    return copyStart;
}


Cell::Cell(int r, int c)
    : row(r), col(c), next(0)
{
 
}


Cell::~Cell()
{

}

