// file: lifedialog.cpp

#include <qdatetime.h>
#include <stdio.h>
#include <stdlib.h>
#include <qfile.h>
#include <qtextstream.h>
#include <qfiledialog.h>
#include <qstringlist.h>
#include <qregexp.h>
#include "lifedialog.h"
#include <qwidget.h>
#include "lifeconstants.h"
#include <iostream.h>

LifeBoard::LifeBoard(QWidget * parent, int c, int r, bool zoom)
    : QWidget(parent) , Life(c, r)
{
    this->parent = parent;
    scale = MAXCOLS/cols < MAXROWS/rows ? MAXCOLS/cols : MAXROWS/rows;
    if(scale > MAXSCALE)
	scale = MAXSCALE;
    if(zoom)
	scale = 5;
    setGeometry(0, 0, cols*scale, rows*scale); 
    setBackgroundMode(NoBackground);
    setEraseColor(white);
    setMouseTracking(true);
    setFocusPolicy(QWidget::ClickFocus);
    offImage = new QPixmap(this->size());
    offPainter = new QPainter(offImage);
    editStart = new Cell(0, 0);
    contourStart = new Cell(0, 0);
    contourLast = contourStart;
    showCursorPosition = false;
    edit = false;
    moving = false;
}

LifeBoard::~LifeBoard() 
{
    delete offPainter;
    offPainter = 0;
    delete offImage;
    offImage = 0;
}

/* Changes the state of cells[i][j] and updates the board 
   accordingly. Note that 'age' must be set to '0' here. */ 


void LifeBoard::setCell(int i, int j)
{
    if ( i < 1 || i > rows || j < 1 || j > cols )
	return;
    if((cells[i][j] = !cells[i][j]))
	population++;
    else
	population--;

    age = 0;
    updateList();
    repaint(false);
}


void LifeBoard::addToEditContour(int row, int col) 
{
    Cell * c = new Cell(row, col);
    contourLast->next = c;
    contourLast = c;
}


void LifeBoard::deleteEditContour()
{
    if(contourStart->next == 0)
	return;
    deleteList(contourStart->next);
    contourStart->next = 0;
    contourLast = contourStart;
}



bool LifeBoard::isInsideContour(int row, int col)
{
    bool up = false, down = false, left = false, right = false;
    Cell * c = contourStart->next;
    while(c != 0) {
	if(c->row == row) {
	    if(c->col < col)
		left = true;
	    else if(c->col > col)
		right = true;
	}
	if(c->col == col) {
	    if(c->row < row)
		down = true;
	    else if(c->row > row)
		up = true;
	}
	c = c->next;
    }
    return (left && right && down && up);
}



void LifeBoard::createEditList()
{
    if(editStart->next != 0) {
	deleteList(editStart->next);
	editStart->next = 0;
    }
    Cell * editLast = editStart;
    Cell * c = start->next;
    while(c != 0) {
	if(isInsideContour(c->row, c->col)) {
	    Cell * cell = new Cell(c->row, c->col);
	    editLast->next = cell;
	    editLast = cell;
	}
	   c = c->next;
    }  
}


void LifeBoard::deleteEditList()
{
    deleteList(editStart->next);
    editStart->next = 0;
}


void LifeBoard::addMovedCells()
{
    if(!moving) {
	createEditList();
	deleteEditContour();
	moving = true;
    }
    else {
	Cell * c = editStart->next;
	while(c != 0) {
	    cells[c->row][c->col] = true;
	    c = c->next;
	}
    }
    updateList();
}
 

void LifeBoard::deleteEditPattern(bool outside)
{
    Cell * c = start->next;
    if(outside) {
	while(c != 0) {
	    if(!isInsideContour(c->row, c->col))
		cells[c->row][c->col] = false;
	    c = c->next;
	} 
    }
    else {
	while(c != 0) {
	    if(isInsideContour(c->row, c->col))
		cells[c->row][c->col] = false;
	    c = c->next;
	}
    }
    updateList();
}


void LifeBoard::moveEditPatternUp()
{
    if(!moving) {
	createEditList();
	deleteEditPattern(false);   // Inside cells are deleted from table.
	moving = true;
    }
    Cell * c = editStart->next;
    while (c != 0) {
	stepUp(c);
	c = c->next;
    }
}

void LifeBoard::moveEditPatternDown()
{
    if(!moving) {
	createEditList();
	deleteEditPattern(false);   // Inside cells are deleted from table.
	moving = true;
    }
    Cell * c = editStart->next;
    while (c != 0) {
	stepDown(c);
	c = c->next;
    }
}
void LifeBoard::moveEditPatternLeft()
{
    if(!moving) {
	createEditList();
	deleteEditPattern(false);   // Inside cells are deleted from table.
	moving = true;
    }
    Cell * c = editStart->next;
    while (c != 0) {
	stepLeft(c);
	c = c->next;
    }
}

void LifeBoard::moveEditPatternRight()
{
    if(!moving) {
	createEditList();
	deleteEditPattern(false);   // Inside cells are deleted from table.
	moving = true;
    }
    Cell * c = editStart->next;
    while (c != 0) {
	stepRight(c);
	c = c->next;
    }
}

/* Position's pixel coordinates are transformed to board's 
   row and column values that are passed to 'setCell'. 
   'boardClicked' is emitted to get the parent to update 
   age and population labels. 
*/

void LifeBoard::mouseHandle( const QPoint &pos )
{
    if(edit)
	return; 
    int i =  pos.y()/scale + 1;
     int j = pos.x()/scale + 1;
    setCell( i, j );
    emit boardClicked();
}

void LifeBoard::mousePressEvent( QMouseEvent *e )
{
    if ( e->button() == QMouseEvent::LeftButton )
	mouseHandle( e->pos() );
}

void LifeBoard::mouseMoveEvent(QMouseEvent * e)
{
    int lastCol = cursorCol;
    int lastRow = cursorRow;
    cursorCol = e->x()/scale + 1;
    cursorRow = e->y()/scale + 1;

    if(edit && (cursorCol != lastCol || cursorRow != lastRow))
	addToEditContour(cursorRow, cursorCol);
    repaint(false);

}

void LifeBoard::mouseDoubleClickEvent(QMouseEvent * e)
{
    if ( e->button() == QMouseEvent::LeftButton ) {
	mouseHandle( e->pos() );
    showCursorPosition = !showCursorPosition;
    }
}

void LifeBoard::leaveEvent(QEvent * e)
{
    (void) e;
    cursorCol = 0;
    cursorRow = 0;
    repaint(false);
}



void LifeBoard::keyPressEvent(QKeyEvent * e)
{
    switch(e->key())
    {
	case Key_Up: 
	    if(edit) {
		if(e->state() & ControlButton)
		    addMovedCells();
		moveEditPatternUp();
	    }
	    else 
		stepUp();
	    break; 
	case Key_Down:
	    if(edit) {
		if(e->state() & ControlButton)
		    addMovedCells();
		moveEditPatternDown();
	    }
	    else 
		stepDown();
	    break;
	case Key_Left:
	    if(edit) {
		if(e->state() & ControlButton)
		    addMovedCells();
		moveEditPatternLeft();
	    }
	    else 
		stepLeft();
	    break;
	case Key_Right:
	    if(edit) {
		if(e->state() & ControlButton)
		    addMovedCells();
		moveEditPatternRight();
	    }
	    else 
		stepRight();
	    break;
	default:
	    break;
    }
    repaint(false);
}


void LifeBoard::showTrail()
{
    for(int i = 0; i <= rows+1; i++) {
	for(int j = 0; j < cols+1; j++) {
	    cells[i][j] = trail[i][j];
	}
    }
    setTrailing(false);
    updateList();
    age = 0;
    for(int i = 0; i <= rows+1; i++) {
	for(int j = 0; j < cols+1; j++) {
	    trail[i][j] = false;
	}
    }
    repaint(false);
} 


/* The bufferimage is first filled with white and then  
   live cells are painted with black. Finally the 
   bufferimage is copied  to the board. */

void LifeBoard::paintEvent( QPaintEvent * e )
{
    (void) e;
    Cell * c = 0; 
    offImage->fill(white);
    if(start != 0)
	c = start->next;
    while(c != 0) {
	offPainter->fillRect((c->col-1)*scale,(c->row-1)*scale,                                              scale, scale, black);
	c = c->next;
    }
    if(edit) {
	if(!moving) {
	    c = contourStart->next;
	    while(c != 0) {
		offPainter->fillRect((c->col-1)*scale,(c->row-1)*scale,                                              scale, scale, red);
		c = c->next;
	    }
	}
	else  {
	    c = editStart->next;
	    while(c != 0) {
		offPainter->fillRect((c->col-1)*scale,(c->row-1)*scale,                                              scale, scale, red);
		c = c->next;
	    }
	}
    }

    if(showCursorPosition && cursorRow > 0) {
	int row = scale*cursorRow, col = scale*cursorCol; 
	int x = cursorCol > cols/2 ? (col - 15*scale) : (col + 3*scale);
	int y = cursorRow > rows/2 ? (row - 5*scale) : (row + 5*scale);  
	QString s;
	offPainter->drawText(x, y,                                                                       s.sprintf("(%d,%d)", cursorRow, cursorCol));
    }
    bitBlt(this, 0, 0, offImage);
}   

void LifeBoard::loadFromASCII(const QString & filename)
{
    QFile file(filename);
    if(!file.open(IO_ReadOnly)) {
	LifeDialog::showMessage("Cannot open the file.");
	return;
    }

    Cell * cell, * first = new Cell(0,0), * prev = first;
    QTextStream stream(&file);
    QString line;
    int row = 0;
    int col = 0;
    int rowShift = 0;
    int colShift = 0;
    population = 0;
    while(!stream.eof()) {
	while((line = stream.readLine().stripWhiteSpace()) != 0) {
	    if(line.isEmpty())
		continue;
	    if(line.startsWith("#Life 1.05"))
		break;
	    else {
		LifeDialog::showMessage("#Life 1.05-line not found.");
		return;
	    }
	}
	while((line = stream.readLine().stripWhiteSpace()) != 0) {
	    if(line.isEmpty())
		continue; 
	    else if(line.startsWith("#")) {
		if(line.at(1) == 'P' || line.at(1) == 'p') {
		    QString str = line.mid(3, line.length()).stripWhiteSpace();
		    col = str.section(' ', 0, 0).toInt();
		    if(col < colShift)
			colShift = col;
		    row = str.section(' ', 1, 1).toInt();
		    if(row < rowShift)
			rowShift = row;
		}
		else  
		    continue;
	    }
	    else {
		row++;

		for(uint i = 0; i < line.length(); i++) {
		    if(line.at(i) == '*') {
			cell = new Cell(row,col+i+1);
			prev->next = cell;
			prev = cell;
		    }
		}
	    }
	}

    if(rowShift < 0 || colShift < 0) {
	cell = first;
	while(cell->next != 0) {
	    cell = cell->next;
	    cell->row = cell->row - rowShift; 
	    cell->col = cell->col - colShift;
	}
    }
    if(loadList(first->next)) {
	LifeDialog::showMessage("The pattern is too big");
	return;
    }
    else { 
	updateList();
	repaint(false);
	}  
    deleteList(first);
    }
}

/* Tries to load a pattern defined in an RLE-file.
   The file extension must be 'RLE',ignoring case.   
*/

void LifeBoard::loadFromRLE(const QString & filename)
{
    QFile file(filename);
    if(!file.open(IO_ReadOnly)) {
	LifeDialog::showMessage("Cannot open the file.");
	return;
    }

    QTextStream stream(&file);
    QString line;
    int colSize = 0, rowSize = 0;
    int rowShift = 0, colShift = 0;
    int row = 1, col = 1;
    population = 0;
    while(!stream.eof()) {
	while((line = stream.readLine()) != 0) {
	    if(line.isEmpty())
		continue;
	    QRegExp re("\\s");
	    int pos;
	    while((pos = re.search(line)) != -1) 
		line = line.remove(pos, 1);
	    re.setPattern("^#");
	    if(re.search(line) != -1)
		continue;
	    if(rowSize == 0) {
		re.setPattern("^x=(\\d+),y=(\\d+)");
		if(re.search(line) != -1) {
		    colSize = re.cap(1).toInt();
		    rowSize = re.cap(2).toInt();
		}
		else {
		    LifeDialog::showMessage("Field size not found");
		    return;
		}
		rowShift = (rows - rowSize)/2;
		colShift = (cols - colSize)/2;
		if(rowSize > rows || colSize > cols) {
		    LifeDialog::showMessage("The pattern is too big");
		    return;
		}
		continue;
	    }
	    re.setPattern("[ac-np-z");
	    while ((pos = re.search(line)) != -1)
	    line = line.replace(pos, 1, "o");  
	    re.setPattern("^\\d+");
	    if(re.search(line) == -1)
		line = line.insert(0, '1');
	    re.setPattern("[bo$]");
	    int start = 0;
	    while((pos = re.search(line, start)) != -1) {
		QString s;
		if(line.at(pos) == 'b')
		    s =  ",b,";
		else if(line.at(pos) == 'o')
		    s = ",o,";
		else if(line.at(pos) == '$')
		    s = ",$,";
		line = line.replace(pos, 1, s);
		start = pos + 2;
	    }

	    QStringList list = QStringList::split(',', line, true);
	    QStringList::Iterator it = list.begin();
	    re.setPattern("\\d+");
	    int count = 0;
	    while(it != list.end()) {
		if(QString::compare(*it, "!") == 0) {
		    file.close();
		    updateList();
		    age = 0;
		    repaint(false);
		    return;
		}
		else if((*it).isEmpty()) 
		    count = 1;
		else if(re.search(*it) != -1)
		    count = (*it).toInt();
		else if(QString::compare(*it, "b") == 0) 
		    col += count;
		else if(QString::compare(*it, "o") == 0) {
		    for(int i = 1; i <= count; i++) {
			cells[row+rowShift][col+colShift] = true;
			population++;
			col++;
		    }
		}
		else if(QString::compare(*it,"$") == 0) {
		    for(int i = 1; i <= count; i++)
			row++;
		    col = 1;
		} 
		it++; 
	    }
	}
    }
    file.close();
    age = 0;
    updateList();
    repaint(false);
}

/* Saves the current generation as an RLE-file. The file extension 
   is 'RLE'.  */

void LifeBoard::saveToRLE(const QString & filename)
{
    QFile file(filename);
    if(!file.open(IO_WriteOnly)) {
	LifeDialog::showMessage("Cannot create the file.");
	return;
    }

    QTextStream stream(&file);
    QString fieldSize = "";
    int minCol = getMinCol(start->next);
    int maxCol = getMaxCol(start->next);
    int minRow = getMinRow(start->next);
    int maxRow = getMaxRow(start->next); 

    fieldSize.sprintf("x = %d, y = %d\n",                                                 maxCol - minCol + 1, maxRow - minRow + 1);
    stream << fieldSize;

    QString line = "";
    int lineCount = 0;
    for(int i = minRow; i <= maxRow; i++) {
	int bCount = 0;
	int oCount = 0; 
	QString s = "";
	for(int j = minCol; j <= maxCol; j++) {
	    if(!cells[i][j]) {
		if(line.length() > 57) {
		    line.append("\n");
		    stream << line;
		    line.setLength(0);
		}
		if(oCount) {  
		    s = s.sprintf("o");
		    if(oCount > 1) 
			s.prepend(QString::number(oCount)); 
		    line.append(s);
		    oCount = 0;
		}
		bCount++;
	    }
	    else {
		if(lineCount) {
		    s =                                                                        lineCount > 1 ? s.sprintf("%d$",lineCount) : s.sprintf("$");
		    line.append(s);
		    lineCount = 0;
		}
		if(bCount == 1) 
		   line.append('b');
		else if(bCount > 1) { 
		   s = s.sprintf("%db",bCount);
		   line.append(s);
		}
		bCount = 0;   
		oCount++;
	    }
	}
	s = "";
	if(oCount == 1) 
	    s = s.sprintf("o");
	else if(oCount > 1) 
	    s = s.sprintf("%do",oCount);
	line.append(s);
	oCount = 0;
	lineCount++;


	if(i == maxRow) {
	    line.append("!\n");
	    stream << line;
	}
    }
    file.close();
}


void LifeBoard::surfaceSlot(int index)
{
    switch(index) {
	case 0:
	    pSurface = &LifeBoard::torus;
	    break;
	case 1:
	    pSurface = &LifeBoard::kleinBottle;
	    break;
	case 2:
	    pSurface = &LifeBoard::ppp;
	    break;
	case 3:
	    pSurface = &LifeBoard::mirror;
	    break;
	case 4:
	    pSurface = &LifeBoard::box;
	    break;
	default:
	    break;

    }
}

void LifeBoard::trailSlot()
{
    setTrailing(true);

} 


MyDialog::MyDialog(LifeDialog * ld)
    : QWidget(ld)
{
    setFixedSize(195, 140);
    setPaletteBackgroundColor(magenta);
    board = ld->board;
    buttonGroup = new QButtonGroup(1, Qt::Horizontal, "Delete", this);
    buttonGroup->setGeometry(20, 10, 150, 80);
    radioButton1 = new QRadioButton("Outside cells", buttonGroup);
    radioButton2 = new QRadioButton("Inside cells", buttonGroup);
    radioButton1->setChecked(true);   
 
    deleteButton = new QPushButton("Delete", this);
    deleteButton->setGeometry(10, 100, 55, 30);
    cancelButton = new QPushButton("Cancel", this);
    cancelButton->setGeometry(70, 100, 55, 30);
    closeButton = new QPushButton("Close", this);
    closeButton->setGeometry(130, 100, 55, 30); 

    connect(deleteButton, SIGNAL(clicked()), this, SLOT(deleteClicked()));
    connect(cancelButton, SIGNAL(clicked()), this, SLOT(cancelClicked()));
    connect(closeButton,  SIGNAL(clicked()), this, SLOT(closeClicked()));
}


void MyDialog::deleteClicked()
{
    bool deleteOutside = radioButton1->isChecked();
    board->deleteEditPattern(deleteOutside);
    board->deleteEditContour();
    board->repaint(false);
} 

void MyDialog::cancelClicked()
{
    board->addMovedCells();
    board->deleteEditList();
    board->setMoving(false);
    board->deleteEditContour();
    board->repaint(false);
}


void MyDialog::closeClicked()
{
    board->addMovedCells();
    board->deleteEditList();
    board->setMoving(false);
    board->deleteEditContour();
    board->repaint(false);
    board->setEdit(false);
    board->setMouseTracking(true);
    hide();
}


LifeTimer::LifeTimer( QWidget * parent ) : QTimer( parent ), interval( 100 )
{

}


void LifeTimer::pause( bool stopIt )
{
    if ( stopIt ) 
	stop();
    else
	start( interval ); 
}

/* Sets the interval (in milliseconds) between consecutive 
   generations. */

void LifeTimer::setSpeed( int speed )
{
    interval = (speed < 380) ? MAXSPEED - speed : 20; 
    if ( isActive() )
	changeInterval( interval );
}


LifeDialog::LifeDialog()
{
    setFixedSize(MAXCOLS + 200, MAXROWS);
    r = DEFROWS;
    c = DEFCOLS;
    board = new LifeBoard(this, c, r);
    board->move(0, 0);
    runButton = new QPushButton("Start", this);
    runButton->setGeometry(410, 5, 55, 30);
    stepButton = new QPushButton("Step", this);
    stepButton->setGeometry(470, 5, 55, 30);
    clearButton = new QPushButton("Clear", this);
    clearButton->setGeometry(530, 5, 55, 30);
    saveBufferButton = new QPushButton("Save\nbuffer", this);
    saveBufferButton->setGeometry(410, 40, 55, 40);
    loadBufferButton = new QPushButton("Load\nbuffer", this);
    loadBufferButton->setGeometry(470, 40, 55, 40);
    prevBufferButton = new QPushButton("Prev.\nbuffer", this);
    prevBufferButton->setGeometry(530, 40, 55, 40); 
    slimButton = new QPushButton("Slim", this);
    slimButton->setGeometry(410, 85, 55, 30);
    trailButton = new QPushButton("Trail", this);
    trailButton->setGeometry(470, 85, 55, 30);
    prevRunButton = new QPushButton("Rerun", this);
    prevRunButton->setGeometry(530, 85, 55, 30);
    loadFileButton = new QPushButton("Load", this);
    loadFileButton->setGeometry(410, 120, 55, 30);
    saveFileButton = new QPushButton("Save", this);
    saveFileButton->setGeometry(470, 120, 55, 30);
    bbcButton = new QPushButton("Bbc", this);
    bbcButton->setGeometry(530, 120, 55, 30);

    val = new QIntValidator(16, 400, this);
    rowLabel = new QLabel("Rows", this);
    rowLabel->setGeometry(410, 155, 40, 20);
    QString rowString = "";
    rowString = rowString.sprintf("%d", this->r);
    rowField = new QLineEdit(rowString, this);
    rowField->setValidator(val);
    rowField->setGeometry(455, 155, 30, 20);
    colLabel = new QLabel("Columns", this);
    colLabel->setGeometry(490, 155, 60, 20);
    QString colString = "";
    colString = colString.sprintf("%d", this->c);
    colField = new QLineEdit(colString, this);
    colField->setValidator(val);
    colField->setGeometry(555, 155, 30, 20);
    resizeButton = new QPushButton("Resize\nboard", this);
    resizeButton->setGeometry(410, 180, 55, 40);
    zoomButton = new QPushButton("Zoom\nin", this);
    zoomButton->setGeometry(470, 180, 55, 40);  
    randomButton = new QPushButton("Random\nseed", this);
    randomButton->setGeometry(530, 180, 55, 40);
    skipButton = new QPushButton("Skip", this);
    skipButton->setGeometry(410, 225, 55, 30);
    symButton = new QPushButton("Sym", this);
    symButton->setGeometry(470, 225, 55, 30);
    editButton = new QPushButton("Edit", this);
    editButton->setGeometry(530, 225, 55, 30);
    editDialog = 0;
    ageLabel1 = new QLabel("Age", this);
    ageLabel1->setGeometry(410, 265, 30, 30);
    ageLabel2 = new QLabel("0", this);
    ageLabel2->setGeometry(445, 265, 40, 30); 
    populationLabel1 = new QLabel("Population", this);
    populationLabel1->setGeometry(485, 265, 65, 30);
    populationLabel2 = new QLabel("0", this);
    populationLabel2->setGeometry(555, 265, 40, 30);
    surfaceLabel = new QLabel("Universe", this);
    surfaceLabel->setGeometry(410, 310, 55, 30);
    QStringList universes;
    universes.append("Torus");
    universes.append("Klein bottle");
    universes.append("PPP");
    universes.append("Mirror");
    universes.append("Box");
    combo = new QComboBox(false, this);
    combo->setGeometry(475,310, 115, 30);
    combo->insertStringList(universes);

    speedLabel = new QLabel("Speed", this);
    speedLabel->setGeometry(410, 360, 40, 30);
    slider = new QSlider(100, 400, 10, 300, QSlider::Horizontal, this);
    slider->setGeometry(460, 360, 100, 30);  
    plusLabel = new QLabel("++", this);
    plusLabel->setGeometry(565, 360, 20, 30);
    timer = new LifeTimer(this);
    buffer = prevBuffer = 0;
    prevRun = 0;
    running = false;
    zooming = false; 

    hSymAct = new QAction("Horizontal",0, this);
    vSymAct = new QAction("Vertical", 0, this);
    hvSymAct = new QAction("Vertical and Horizontal", 0, this);
    pdSymAct = new QAction("Diagonal", 0, this);
    mdSymAct = new QAction("Antidiagonal", 0, this);
    QString degree = "\u00b0";
    r90SymAct = new QAction("Rotational 90" + degree, 0, this);
    r180SymAct = new QAction("Rotational 180" + degree, 0, this);
    r270SymAct = new QAction("Rotational 270" + degree, 0, this);

    connect(hSymAct, SIGNAL(activated()), this, SLOT(hSymSlot()));
    connect(vSymAct, SIGNAL(activated()), this, SLOT(vSymSlot()));
    connect(hvSymAct, SIGNAL(activated()), this, SLOT(hvSymSlot()));
    connect(pdSymAct, SIGNAL(activated()), this, SLOT(pdSymSlot()));
    connect(mdSymAct, SIGNAL(activated()), this, SLOT(mdSymSlot()));
    connect(r90SymAct, SIGNAL(activated()), this, SLOT(r90SymSlot()));
    connect(r180SymAct, SIGNAL(activated()), this, SLOT(r180SymSlot()));
    connect(r270SymAct, SIGNAL(activated()), this, SLOT(r270SymSlot()));


    connect(runButton, SIGNAL(clicked()), this, SLOT(startSlot()));
    connect(stepButton, SIGNAL(clicked()), this, SLOT(stepSlot()));
    connect(clearButton, SIGNAL(clicked()), this, SLOT(clearSlot()));
    connect( slider, SIGNAL(valueChanged(int)),
	     timer,  SLOT(setSpeed(int)) );
    connect(saveBufferButton, SIGNAL(clicked()), this, SLOT(saveBufSlot()));
    connect(loadBufferButton, SIGNAL(clicked()), this, SLOT(loadBufSlot())); 
    connect(prevBufferButton, SIGNAL(clicked()), this, SLOT(prevBufSlot()));
//    connect(rowField, SIGNAL(returnPressed()), this, SLOT(rowSlot()));
//    connect(colField, SIGNAL(returnPressed()), this, SLOT(colSlot()));
    connect(resizeButton, SIGNAL(clicked()), this, SLOT(resizeSlot()));
    connect(timer, SIGNAL(timeout()), this, SLOT(runSlot()));
    connect(timer, SIGNAL(timeout()), this, SLOT(ageSlot()));
    connect(board, SIGNAL(boardClicked()), this, SLOT(ageSlot()));
    connect(randomButton, SIGNAL(clicked()), this, SLOT(randomSlot()));
    connect(zoomButton, SIGNAL(clicked()), this, SLOT(zoomSlot()));
    connect(loadFileButton, SIGNAL(clicked()), this, SLOT(loadFileSlot()));
    connect(saveFileButton, SIGNAL(clicked()), this, SLOT(saveFileSlot()));
    connect(combo, SIGNAL(activated(int)), board, SLOT(surfaceSlot(int)));
    connect(bbcButton, SIGNAL(clicked()), this, SLOT(bbcSlot()));
    connect(slimButton, SIGNAL(clicked()), this, SLOT(slimSlot()));
    connect(trailButton, SIGNAL(clicked()), board, SLOT(trailSlot()));
    connect(prevRunButton, SIGNAL(clicked()), this, SLOT(prevRunSlot())); 
    connect(skipButton,SIGNAL(clicked()), this, SLOT(skipSlot()));
    connect(symButton, SIGNAL(clicked()), this, SLOT(symSlot())); 
    connect(editButton, SIGNAL(clicked()), this, SLOT(editSlot()));
}


LifeDialog::~LifeDialog()
{
    if(buffer != 0)
	deleteBuffer(buffer);
    if(prevBuffer != 0)
	deleteBuffer(prevBuffer);
}


void LifeDialog::deleteBuffer(Cell * buf)
{
    Cell * prev;
    Cell * b = buf;
    while(b != 0) {
	prev = b;
	b = b->next;
	delete(prev);
    }
}


void LifeDialog::ageSlot()
{
    ageLabel2->setNum(board->getAge());
    populationLabel2->setNum(board->getPopulation());
} 



void LifeDialog::startSlot()
{

    running = !running;
    timer->pause(!running);
    if(running) {
	scp = board->getSCP();
	board->setSCP(false);
	runButton->setText("Stop");
	if(prevRun != 0)
	    deleteBuffer(prevRun);
	prevRun = board->toList();
    }
    else {
	board->setSCP(scp); 
	runButton->setText("Start");
	if(board->getTrailing()) {
	    board->doTrail();
	    board->showTrail();
	    ageSlot();
	    board->setFocus();
	}
    }
}

void LifeDialog::stepSlot()
{
    if(!running) {
	board->nextGen();
	board->repaint(false);
	ageSlot();
	board->setFocus();
    }
}


void LifeDialog::clearSlot()
{
    if(!running) {
	board->clear();
	board->repaint(false);
	ageSlot();
	board->setFocus();
    }
}


void LifeDialog::saveBufSlot()
{
    if(!running) {
	if(prevBuffer != 0) 
	    deleteBuffer(prevBuffer);
	prevBuffer = buffer;
	buffer = board->toList();
	board->setFocus();
    }
}


void LifeDialog::loadBufSlot()
{
    if(buffer == 0 || running)
	return;
    if(board->loadList(buffer))
	showMessage("Pattern is too big");
    else {
	board->updateList();
	board->repaint(false);
	ageSlot();
	board->setFocus();
    }
}

void LifeDialog::prevBufSlot()
{
    if(prevBuffer == 0 || running)
	return;
    if(board->loadList(prevBuffer))
	showMessage("Pattern is too big");
    else {
	board->updateList();
	board->repaint(false);
	ageSlot();
	board->setFocus();
    }
}


void LifeDialog::rowSlot()
{
    this->r = rowField->text().toInt();
}


void LifeDialog::colSlot()
{
    this->c = colField->text().toInt();
}


void LifeDialog::randomSlot()
{
    if(!running) {
	board->randomSeed();
	board->repaint(false);
	ageSlot();
    }
}

void LifeDialog::runSlot() 
{
    board->nextGen();
    board->repaint(false);
}


void LifeDialog::resizeSlot()
{
    if(running)
	return;
    this->r = rowField->text().toInt();
    this->c = colField->text().toInt();
    delete board;
    board = new LifeBoard(this, this->c, this->r); 
    board->move(0, 0);
    board->show();
    connect(board, SIGNAL(boardClicked()), this, SLOT(ageSlot()));
    connect(trailButton, SIGNAL(clicked()), board, SLOT(trailSlot()));
    connect(combo, SIGNAL(activated(int)), board, SLOT(surfaceSlot(int)));
    board->repaint(false);
}


void LifeDialog::zoomSlot()
{
    Cell * start = 0;
    zooming = !zooming;
    int r = board->getRows();
    int c = board->getCols();
    if(board->getPopulation() > 0)
	start = board->toList();
    delete board;
    board = new LifeBoard(this, c, r, zooming);

    if(zooming) {
	scrollView = new QScrollView(this);
	scrollView->setGeometry(0, 0, MAXCOLS, MAXROWS);
	scrollView->addChild(board);
	scrollView->center((board->width())/2, (board->height())/2);
	scrollView->show();
	zoomButton->setText("Zoom\nout");
    }
    else {
	scrollView->removeChild(board);
	delete scrollView;
	scrollView = 0;
	zoomButton->setText("Zoom\nin");
    }
    board->loadList(start);
    board->show();
    connect(board, SIGNAL(boardClicked()), this, SLOT(ageSlot()));

}


void LifeDialog::loadFileSlot()
{
    if(running)
	return;
//    board->releaseKeyboard();
    QString fileName;
    QString types("*.RLE *.LIFE *.LIF *.rle *.life *.lif");

    if((fileName = QFileDialog::getOpenFileName(                                                    "patterns",types, board )) ==      QString::null) 
	return;
    if(fileName.endsWith(".RLE",false)) {
	board->loadFromRLE(fileName);
	ageSlot();
    }
    else if(fileName.endsWith(".LIFE",false) ||                                                                      fileName.endsWith(".LIF",false)) {
	board->loadFromASCII(fileName);
	ageSlot();
    }
    board->setFocus();
}

void LifeDialog::saveFileSlot()
{
    if(running)
	return;
//    board->releaseKeyboard();
    QString fileName;
    if((fileName = QFileDialog::getSaveFileName(                                                    "patterns","*.RLE", board)) !=    QString::null) {
    board->saveToRLE(fileName);
    }
    board->setFocus();
}

void LifeDialog::bbcSlot()
{
    if(running)
	return;
    board->createBBC();
    board->repaint(false);
    ageSlot();
}

void LifeDialog::slimSlot()
{
    if(running)
	return;
    board->createSlim();
    board->repaint(false);
    ageSlot();
}


void LifeDialog::prevRunSlot()
{  
    if(board->loadList(prevRun))
	showMessage("The pattern is too big");
    else {
	board->updateList();
	board->repaint(false);
    }
}




void LifeDialog::skipSlot()
{
    bool ok = false;
    int ret = QInputDialog::getInteger("Skipping",                                            "Enter number of generations", 0, 1, 200000, 1, &ok);
    QString s = "";
    s = s.sprintf("Skipping %d generations", ret);

    if(ok) {
	Cell * bup = board->toList();
	QProgressDialog progress(s, "Abort skipping",                                                            ret, this, "progress", true);
	for(int i = 0; i < ret; i++) {
	    if(!(i % 10))
		progress.setProgress(i);
//          qApp->processEvents();
	    if(progress.wasCanceled()) {
		board->clear();
		board->loadList(bup);
		board->updateList();
		break;
	    }
	    board->nextGen();
	}
	progress.setProgress(ret);
	board->repaint(false);
	ageSlot();
	deleteBuffer(bup);
    }
}                                              

void LifeDialog::symSlot()
{
    symMenu = new QPopupMenu(this);
    hSymAct->addTo(symMenu);
    vSymAct->addTo(symMenu);
    hvSymAct->addTo(symMenu);
    pdSymAct->addTo(symMenu);
    mdSymAct->addTo(symMenu);
    r90SymAct->addTo(symMenu);
    r180SymAct->addTo(symMenu);
    r270SymAct->addTo(symMenu);
    symMenu->exec(mapToGlobal(symButton->pos()));
}

void LifeDialog::hSymSlot()
{
    board->hSymmetrize();
    board->repaint(false);
    delete symMenu;
}

void LifeDialog::vSymSlot()
{
    board->vSymmetrize();
    board->repaint(false);
    delete symMenu;
}


void LifeDialog::hvSymSlot()
{
    board->hSymmetrize();
    board->vSymmetrize();
    board->repaint(false);
    delete symMenu;
}


void LifeDialog::pdSymSlot()
{
    if(!board->pdSymmetrize()) 
	showMessage("The operation cannot be performed!");
    else 
	board->repaint(false);
    delete symMenu;
}


void LifeDialog::mdSymSlot()
{
    if(!board->mdSymmetrize()) 
	showMessage("The operation cannot be performed!");
    else 
	board->repaint(false);
    delete symMenu;
}


void LifeDialog::r90SymSlot()
{
    if(!board->r90Transform()) 
	showMessage("The operation cannot be performed!");
    else 
	board->repaint(false);
    delete symMenu;
}


void LifeDialog::r180SymSlot()
{
    board->r180Symmetrize(); 
    board->repaint(false);
    delete symMenu;
}

void LifeDialog::r270SymSlot()
{
    if(!board->r270Transform()) 
	showMessage("The operation cannot be performed!");
    else 
	board->repaint(false);
    delete symMenu;
}

void LifeDialog::editSlot()
{
    board->setEdit(true);
    board->setMouseTracking(false);

    if(!editDialog) {
    editDialog = new MyDialog(this);
    editDialog->move(405,200);;
    }
//    QPoint p = editDialog->pos();
//    cout << p.x() << " " << p.y() << endl;
    editDialog->show();

}

void LifeDialog::showMessage(const QString & msg)
{
 (void)  QMessageBox::information(0, "Qlife", msg,                                        QMessageBox::Ok | QMessageBox::Default);
}



                            

